/**  
    The network "middle-tier" of our multi-client Go game -- think of
    them as "ears" of the server, or ports on the hub.  The server's
    thread can't do anything but wait for new socket connections.
    Therefore the server creates *these* threads for the sole purpose
    of listening to incoming messages from each MessagePasser client.
*/


import java.io.*;
import java.net.*;
import java.awt.*;
import java.awt.event.*;


public class ClientConnector extends Thread {

  /** Each instance has two buffered streams for communication 
      with a MessagePasser (/GoGame) object */

  public MessageServer my_server;
  public GoPlayer my_player;
  private DataInputStream my_input;
  private PrintWriter my_output;
  private Socket my_sock;
  private int my_ID;
    
    
  /** Constructor sets up its I/O streams over the socket; Then it
      sends off the server's GoBoard, and receives back the client's
      GoPlayer.  The miracle of object serialization, eh?  */

  public ClientConnector(Socket my_sock, MessageServer server, int ID) 
    throws IOException 
  {
    this.super();
    this.my_server = server;
    this.my_sock = my_sock;
    this.my_ID = ID;
    my_input = new DataInputStream(my_sock.getInputStream());
    my_output = new PrintWriter(my_sock.getOutputStream());
    my_player = new GoPlayer("nobody", 0, false);

  }

  
  /** Try to handshake with client, return success or failure */

  public boolean do_handshake()
  {
    String reply;

    // Do the handshake
    send_message("BKGO-Server-version-" + my_server.version.version_number);
    reply = readLine(my_input);
    if (reply == null) return false;
    // And here, someday, we will bother to care about the BKGO-client's
    // version-response, and do protocol negotiations.
    send_message("OK");
    reply = readLine(my_input);
    if (reply == null) return false;

    if (reply.equalsIgnoreCase("status")) {
      GoPlayer gp;
      String player_list, turn = "Nobody's";
      if (my_server.board.to_move() == 1) { turn = "Black's"; }
      if (my_server.board.to_move() == 2) { turn = "White's"; }

      player_list = "Server Name: " + my_server.myname + "\n";
      player_list = player_list + "Moves made: " 
	+ my_server.board.number_of_moves() + "\n";
      player_list = player_list + 
	"Black score: " + my_server.board.black_score() + "\n";
      player_list = player_list + 
	"White score: " + my_server.board.white_score() + "\n";
      player_list = player_list + turn + " turn.\n\n";

      player_list = player_list + "Players currently connected:\n";
      for (int j = 0; j < my_server.players.size(); j++) // list all players
      { 
        gp = (GoPlayer) my_server.players.elementAt(j);
	player_list = player_list + gp.toPrettyString() + "\n";
      }
      if (my_server.players.size() == 0)
	player_list = player_list + "none.";
      send_message(player_list);
      return false;  // now kill me
    }
    else if (reply.equalsIgnoreCase("join")) { 
      // Request the client's GoPlayer object!
      send_message("GimmeGoPlayer"); 
      // And wait for client to return a GoPlayer string:
      System.out.println("CC" + my_ID + ": About to receive GoPlayer...");
      String message = readLine(my_input);
      if (message.startsWith("Player: ")) {
	String fields = message.substring(8);
	my_player.parseString(fields);
	// After receiving the GoPlayer, tell the server!
	my_server.add_player(this);
	return true; // start my run() thread
      }
      else {
	send_message("Protocol error.");
	System.out.println("CC" + my_ID + ": protocol error!");
	return false;
      }
    }
    else if (reply.equalsIgnoreCase("shutdown")) {
      if ( (my_server.blackpassword.equals("none")) &&
	   (my_server.whitepassword.equals("none")) )
      {
	// go ahead and shutdown
	send_message("server terminated");
	my_server.broadcast("Dialog: SERVER HAS BEEN TERMINATED");
	my_server.quit();
	return false;  // now kill me
      }
      else {
	String password = fetch_password(false);
	if (password == null) {  // couldn't fetch proper password
	  send_message("Protocol error.");
	  System.out.println("CC" + my_ID + ": protocol error!");
	  return false; // now kill me
	}
	// otherwise, check it against the two owner passwords
	if ( (my_server.blackpassword.equals(password)) ||
	     (my_server.whitepassword.equals(password)) )
	{
	  // go ahead and shutdown
	  send_message("server terminated");
	  my_server.broadcast("Dialog: SERVER HAS BEEN TERMINATED");
	  my_server.quit();
	  return false;  // now kill me
	}
	else {
	  send_message("password denied");
	  return false;
	}
      }
    }
      
    else {  // unknown handshake
      send_message("Protocol error.");
      System.out.println("CC" + my_ID + ": protocol error!");
      return false;  // now kill me
    }

  }


  /** Send the a GoBoard to our client */

  public void send_board(GoBoard new_board)
  {
    send_message("Board: ");  // tell client it's coming.
    System.out.println("CC" + my_ID + ": About to send out new GoBoard...");
    String SGFmessage = new_board.toString();
    send_message(SGFmessage);
    send_message("EOF");
    System.out.println("CC" + my_ID + ": Sent new board to client.");
  }

  /** Retrieve a GoBoard from our client;  store it in the server too. */

  public GoBoard receive_board()
  {
    GoBoard new_gb = new GoBoard(19);
    String message_part = null;
    StringBuffer SGFbuf = new StringBuffer();
    new_gb = new GoBoard(19);
    SGFnode NewTree = new SGFnode();

    System.out.println("CC" + my_ID + ": About to get my client's board...");
    message_part = readLine(my_input);

    // Keep reading lines from the SGF stream
    while (! message_part.equals("EOF")) {
      SGFbuf.append(message_part);
      message_part = readLine(my_input);
    }
      
    // Call the parser to build our tree
    NewTree.fromString(SGFbuf.toString());
    // Install our new blank board
    my_server.board = new_gb;
    my_server.board.GameTree = NewTree;
    // Execute the new tree
    NewTree.execute(my_server.board);

    System.out.println("CC" + my_ID + ": storing this board within server.");
  
    return my_server.board;
  }


  /** Sends a text message to the MessagePasser/GoGame; called by
      MessageServer.broadcast() */
    
  public void send_message(String message) {
    my_output.println(message);
    my_output.flush();
    if (my_output.checkError()) {
      System.out.println("CC"+my_ID+": Error in socket write;  killing self.");
      die(true);
    }
  }

  /** Thanks to Kevin Kelley <kelley@ruralnet.net> for this
      function.  The only non-deprecated readLine() seems to be in
      BufferedReader, which royally messes up our stream when doing
      serialization.  Hence, a home-made readLine().  */

  public String readLine(InputStream in)
  {
    StringBuffer buffer = new StringBuffer(300);
    char ch = 0;
    try {
      while ((ch = (char) in.read()) != -1) { // break on EOF
        if ((ch == '\n') || (ch == '\r')) {
          if (ch == '\r') // if line ends with carriage return,
            in.read();  // read trailing newline and discard it
          break;
        }
        if (buffer.length() > 995) {
          System.out.println
            ("CC" + my_ID +": readLine() received > 995 chars. Killing self!");
	  if (my_server.players.size() == 0)
	    die(false);
	  else
	    die(true);
	  return null;
        }
        else 
          buffer.append(ch);
      }
      // return null to indicate EOF; otherwise return the line.
      if (ch == -1 && buffer.length() == 0) 
        return null;
      else
        return buffer.toString();
    }
    catch (IOException e) {
      System.out.println("CC" + my_ID + ": readLine() IOException,  Killing self.");
      if (my_server.players.size() == 0)
	die(false);
      else
	die(true);
    }
    catch (OutOfMemoryError e) {
      System.out.println("CC" + my_ID + ": readLine OutOfMemory.  Killing self.");
      if (my_server.players.size() == 0)
	die(false);
      else
	die(true);
    }
    return null;
  }


  /** Kill myself and the player attached to me! */

  public void die(boolean remove_player) {
    // If the server has already registered the player:
    if (remove_player) {
      my_server.remove_player(my_player.IDnum);
      this.stop();  // stop trying to readLine();
    }
    // No matter what, close all my streams
    try {
      my_input.close();
      my_output.close();
      my_sock.close();
      System.out.println("CC" + my_ID + ": closed my socket.");
    }  
    catch (IOException e) {
      System.out.println("CC" + my_ID + ": IOException on die(): " + e);
    }
  }


  /** This routine blocks for a password from the client.

      It *is* thread-safe, since our run()-readLine() can't possibly 
      be going at the same time, because the *same* ClientConnector
      will always be the very thread executing this routine!
  */

  public String fetch_password(boolean new_p)
  {
    if (new_p)
      send_message("GimmeNewPassword");
    else
      send_message("GimmePassword");

    System.out.println("CC" + my_ID + ": waiting for password...");
    // wait for the password to arrive:
    String message = readLine(my_input);
 
    if (message == null) {
      return null;
    }

    if (message.startsWith("Password: "))
    {
      String password = message.substring(10);
      return password;
    }
    else return null;
  }


  /** Overriding Thread's run() method; called when object is
      start()ed.

      This routine parses messages received from the client and
      updates MessageServer's GoBoard -- exactly the way a client
      should.  Then, it broadcasts the same message to all the clients
      so they can do their own board updates.  */
    

  public void run() {

    while (true) {

      // wait for a MessagePasser message indefinitely
      String message = readLine(my_input);

      if (message == null) {
        System.out.println
          ("CC" + my_ID + ":  Received garbage!  That's it, I'm outta here!");
        die(true);
      }

      // then parse the message!
      if (message.startsWith("Quitting: ")) {
        System.out.println("CC" + my_ID + ": Player is leaving, and so am I.");
        die(true);
      }

      else if (message.startsWith("Place: ")) {
	// add a stone to the server's GoBoard
	String coords = message.substring(7);
	my_server.received_move(coords);
        my_server.broadcast(message);  // valid message	      
      }

      else if (message.startsWith("Message: ")) {
	String real_message = message.substring(9);
	my_server.board.add_comment_node(real_message);
        my_server.broadcast(message);  // valid message	      
      }

      else if (message.startsWith("Pass")) {
	my_server.board.pass();
        my_server.broadcast(message);  // valid message	      
      }

      else if (message.startsWith("SendingBoard")) {
	// grab the GoBoard
	GoBoard new_board = receive_board();
	my_server.broadcast_board(new_board);
      }

      else if (message.startsWith("NewBoard")) {
	// Client wants us to give its board to everyone.
	send_message("GimmeBoard");
      }

      else if (message.startsWith("SetName: ")) {
	String potential_name = message.substring(9);
	my_server.myname = potential_name;
	my_server.broadcast("Dialog: Server has been named '"
			      + potential_name +"'.");
      }

      // If we have received unparseable garbage, tell the server
      // to kill the player!  (This is the same code in "Quitting:"
      // above.)
      else {   
        System.out.println
          ("CC" + my_ID + ":  Received garbage!  That's it, I'm outta here!");
        die(true);
      }
    }
  }
    
}


//////////////////////////////////////////////////////////
// Don't think, just hit `y':
//
// local variables:
// eval: (load-file "bkgomagic.el")
// end:
//////////////////////////////////////////////////////////
