/** The GoBoard class holds the state of the game; GoCanvas displays it. */

import java.io.*;
import java.awt.*;
import java.applet.*;
import java.text.*;
import java.util.*;



class GoBoard implements Serializable
{
  // This is new and tricky here.  When "executing" a GameTree,
  // GoBoard may need to access information in its parent GoGame (such
  // as the chat window).  On the other hand, if GoBoard belongs to a
  // MessageServer instead, there's no need.  Therefore, this "game"
  // variable will either be null (when part of MessageServer) or it
  // will actually point to a legitimate GoGame parent wrapper.  On
  // this same note, it's important for MessageServer and GoGame to
  // call one of two *different* constructors!

  public GoGame game = null;

  // Used for SGF board coordinates - FF[4] handles boards up to 52x52.
  String letters[] = {"a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k",
		    "l", "m", "n", "o", "p", "q", "r", "s", "t", "u", "v",
		    "w", "x", "y", "z", "A", "B", "C", "D", "E", "F", "G",
		    "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q", "R",
		    "S", "T", "U", "V", "W", "X", "Y", "Z" };
  
  // Compass directions from a given point.  See get_point(pt, dir).
  static final int UP = 1, DOWN = 2, LEFT = 3, RIGHT = 4;
  
  // Whose move is it?  Black starts by default.
  // todo: make accessor method for this
  public int to_move = GoPoint.BLACK;
  
  // Added by Ben:  keep track of how many stones are captured.
  public int blackscore = 0;
  public int whitescore = 0;
  
  // POWER-SHIFT: some meta-game information which *used* to reside in
  // either the BKGo server or GoGame wrapper-class has now been moved
  // into GoBoard.  This gives GoBoard the ability to construct a Root
  // SGFnode with good meta-game properties, and serves as a nice
  // "common ground" for both client and server sides.  (Remember,
  // both client and server keep GoBoards of their own.)
  
  public GoVersion version;
  public String blackPlayerName = "nobody";  // set via bkgo protocol
  public String whitePlayerName = "nobody";  // set via bkgo protocol
  
  // This used to be called "moves"; it now is the "root" node of our
  // SGF GameTree!
  public SGFnode GameTree;
  
  // For ultraconvenience, we also keep a pointer to the "last" node
  // on our GameTree.
  public SGFnode lastNode;

  // Holds the state of the game at any given time.  This is different
  // from the history of the game.  For example if a stone is
  // captured, it will be colored empty in `board', but the move that
  // placed it previously would still be a part of the game's record.
  private Vector board;

  // How big is the square board?
  private int size = -1;

  // Used to place the famous red dot:
  GoPoint last_point = null;

  // todo: I think this might belong inside a method, actually:
  private static final int empty_gid = 0;
  private int base_gid;

  // Constructor to be called by MessageServer:
  public GoBoard(int size) {
    version = new GoVersion(); // initialize our "version" object
    size(size);  // set up a blank "board" vector of specific size
    init_gametree();  // create a generic GameTree root node
  }

  // Constructor to be called by GoGame:
  public GoBoard(GoGame parent_game, int size) {
    game = parent_game;  // hold a pointer to our parent GoGame
    version = new GoVersion(); // initialize our "version" object
    size(size);  // set up a blank "board" vector of specific size
    init_gametree();  // create a generic GameTree root node
  }

  // Returns the size of the board.
  // Note that the vector is 1 longer than this, since we're using
  // 1-based indexing in it.
  public int size() {
    return size;
  }

  // Sets up new size, new vectors, returns new size.
  public int size(int new_size)
  {
    /*    if (this.size == new_size)
	  return this.size();*/

    // else

    // Resizing is kind of like "init" for GoBoards.  Since there's no
    // way to expand the board size for a game in progress, resizing
    // effectively starts a new game.

    this.size = new_size;
    board = new Vector(this.size() + 1);  // incr for 1-based indexing
    board.setSize(this.size() + 1);
    for (int i = 1; i <= this.size(); i++)
    {
      Vector v = new Vector(this.size() + 1);
      v.setSize(this.size() + 1);  // todo: necessary?
      for (int j = 1; j <= this.size(); j++)
      {
        GoPoint p = new GoPoint (this, i, j, GoPoint.EMPTY, 0);
	p.previous_color(GoPoint.EMPTY);
        v.setElementAt(p, j);
      }
        
      board.setElementAt(v, i);
    }
    
    return this.size();
  }


  // Sets up a generic properties for a new root SGFnode.

  private void init_gametree() 
  {
    // create our root SGFnode and fill in basic properties
    GameTree = new SGFnode();
    GameTree.head_p = true;  // mark this as the top of a tree (i.e. "(")
    GameTree.set_prop_literal("FF", "4");
    GameTree.set_prop_literal("GM", "1");
    GameTree.set_prop_literal("SZ", Integer.toString(this.size()));
    GameTree.set_prop_literal("AP", "BKGo:" + version.toString());
    GameTree.set_prop_literal("CP", "GNU GPLv2");
    GameTree.set_prop_literal("GN", "default.sgf");  // overwritten by SaveGame()
    GameTree.set_prop_literal("PW", whitePlayerName); // set via bkgo protocol
    GameTree.set_prop_literal("PB", blackPlayerName); // set via bkgo protocol
    GameTree.set_prop_literal("MN", "0");          // the root node is the 0th move.
    Date startdate = new Date();
    SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd");
    GameTree.set_prop_literal("DT", formatter.format(startdate));
    
    // And our lastNode points to our root node, of course.
    lastNode = GameTree;
  }
    

  
  public GoPoint last_point() {
    return last_point;
  }
  
  public int to_move() {
    return to_move;
  }
  
  /** The elegant formula is:
      Captured black pieces =   (number of black stone-placements)
                              - (number of stones on board)
                              - (number of black handicap stones)
  */

  private int captured_count(int color)
  {
    int stone_placed_count = 0, on_board_count = 0;
    SGFnode currentnode = GameTree;
    String col = null;
    GoPoint testpoint = null;
    
    if (color == GoPoint.WHITE)
      col = "W";
    if (color == GoPoint.BLACK)
      col = "B";

    // Traverse our main-line of GameTree, tallying up stone placements
    // of a particular color (we don't do handicap stones yet!)
    while (true)
    {
      // look for stone-placement property (do *not* count a "pass" move!)
      if ( (currentnode.properties.containsKey(col)) &&
	   (currentnode.get_prop(col) != "") )
	stone_placed_count++;
      // and move down the tree
      if (currentnode.children.size() == 0)
	break;
      else
	currentnode = (SGFnode) currentnode.children.firstElement();
    }
    
    // Figure out how many stones of said color are actually *on* the board!
    for (int i = 1; i <= size; i++)
      for (int j = 1; j <= size ; j++) 
      {
	testpoint = get_point(i, j);
	if (testpoint.color() == color) 
	  on_board_count++;
      }
    
    return (stone_placed_count - on_board_count);
  }
  
  public int black_score() {
    return captured_count(GoPoint.WHITE);
  }

  public int white_score() {
    return captured_count(GoPoint.BLACK);
  }

  public int number_of_moves() {
    // ahh, ain't this elegant?
    String num_moves = sgf_nextMoveNumber(lastNode);
    // report one less than this, since we received the "next" number...
    return Integer.parseInt(num_moves) - 1;
  }

  public void set_blackPlayerName(String name) {
    blackPlayerName = name;
    GameTree.set_prop_safely("PB", blackPlayerName); // set via bkgo protocol
  }

  public void set_whitePlayerName(String name) {
    whitePlayerName = name;
    GameTree.set_prop_safely("PW", whitePlayerName); // set via bkgo protocol
  }

  public GoPoint get_point(Point p)
  {
    Vector v;
    GoPoint pt;
    
    // 1-based indexing means 0 is a legal vector address for Java,
    // but not for our application, so guard against it:
    if ((p.x < 1) || (p.y < 1))
      return null;

    v = (Vector) board.elementAt(p.x);
    pt = (GoPoint) v.elementAt(p.y);

    return pt;
  }

  public GoPoint get_point(int x, int y)
  {
    Vector v;
    GoPoint pt;
    
    // 1-based indexing means 0 is a legal vector address for Java,
    // but not for our application, so guard against it:
    if ((x < 1) || (y < 1))
      return null;

    v = (Vector) board.elementAt(x);
    pt = (GoPoint) v.elementAt(y);

    return pt;
  }

  /**  Returns a point near p, or null if it would be off the board. */

  public GoPoint get_point(Point p, int direction)
  {
    int x, y;

    try
    {
      x = p.x;
      y = p.y;

      switch (direction)
      {
      case GoBoard.UP:
        y--;
        break;
      case GoBoard.DOWN:
        y++;
        break;
      case GoBoard.LEFT:
        x--;
        break;
      case GoBoard.RIGHT:
        x++;
        break;
      default:
        // todo: sanity check direction here?
      }
      
      return get_point(new Point(x, y));
    }
    catch (IndexOutOfBoundsException e)
    {
      return null;
    }
  }
  
  private void remove_group (int gid)
  {
    for (int x = 1; x <= size(); x++) {
      for (int y = 1; y <= size(); y++) 
      {
        Point p = new Point (x, y);
        GoPoint pt = get_point (p);
        if (pt.gid_is(gid)) {
          pt.remove_stone(); // make GoPoint EMPTY color
        }
      }
    }
  }

  /** Has this stone's group any liberties? */
  private boolean reaches_oxygen_p (int gid)
  {
    for (int x = 1; x <= size(); x++) {
      for (int y = 1; y <= size(); y++) 
      {
        Point p = new Point (x, y);
        GoPoint pt = get_point (p);
        if (pt.gid_is(gid) && (pt.liberties() > 0))
          return true;
      }
    }
    return false;
  }

  private boolean reaches_oxygen_p (GoPoint pt) {
    return reaches_oxygen_p(pt.gid());
  }

  /** Helper for remove_dead_groups() and dead_groups_p().
   *  Returns true if any got removed / would get removed.
   */
  private boolean maybe_remove_dead_groups(int color, boolean do_it)
  {
    boolean any_dead_groups_p = false;

    // todo: if efficiency is ever a concern, here is one place to
    // look!  This code may check many spots multiple times, depending
    // on the connectivity of the board.

    for (int x = 1; x <= size(); x++) {
      for (int y = 1; y <= size(); y++) 
      {
        Point p = new Point (x, y);
        GoPoint pt = get_point(p);

        if (! pt.color_is(color))
          continue;
        if (reaches_oxygen_p (pt))
          continue;
        // else
        any_dead_groups_p = true;
        if (do_it)
          remove_group(pt.gid());
      }
    }

    // Sanitize, although caller probably will too:
    establish_groups();
    
    return any_dead_groups_p;
  }

  private boolean remove_dead_groups(int color) {
    return maybe_remove_dead_groups (color, true);
  }

  private boolean dead_groups_p(int color) {
    return maybe_remove_dead_groups (color, false);
  }

  /**  Helpers for move(): */

  /** Helper for establish_groups(), which is a helper for move().
   * Establishes a group here, if there's a stone to start from and it
   * hasn't already been visited.  Just does a recursive floodfill.
   *
   * Of course, someday the orientation of the board labels might
   * change, but that probably won't affect this code, which doesn't
   * care how things are displayed on the screen.  "Upper left" here
   * just means 1,1 and "lower right" means 19,19, in the board
   * vector.
   */
  private boolean fill_from_point (GoPoint pt, int color, int this_gid)
  {
    // Most likely we'll just get outta town:
    if ((pt == null)
        || (! pt.color_is(color))
        || (! pt.gid_is(empty_gid)))
      return false;

    // else

    // Set the group here:
    pt.gid (this_gid);

    // Recurse on each of the compass points:
    fill_from_point (get_point(pt, UP),    color, this_gid);
    fill_from_point (get_point(pt, DOWN),  color, this_gid);
    fill_from_point (get_point(pt, LEFT),  color, this_gid);
    fill_from_point (get_point(pt, RIGHT), color, this_gid);

    return true;
  }

  /** Mark-and-sweep group setting:
   *
   * Does a floodfill algorithm, setting the group_id of each stone
   * according to what group it's a member of (the exact gid doesn't
   * matter, the only thing that matters is that all the stones in a
   * given group have the _same_ gid).
   *
   * So: first set everyone's gid to zero.  Then loop-a-doop-doop, not
   * worrying about efficiency, doing what the above paragraph says.  
   */
  private void establish_groups ()
  {
    // The "mark" stage of a mark-and-sweep:
    //
    // todo: When all this is stable, abstract out get_point() so the
    // temporary Point here isn't necessary.  But do wait until
    // stability, because I remember a *nightmare* of inexplicable
    // behaviors the last time I tried abstracting the coordinates for
    // get_point()!  Once burned, thrice cautious...

    for (int x = 1; x <= size(); x++) {
      for (int y = 1; y <= size(); y++) 
      {
        Point p = new Point (x, y);
        GoPoint gp = get_point (p);
        gp.gid (empty_gid);
      }
    }

    // The "sweep" stage:

    base_gid = (empty_gid + 1);  // a good inital value
    for (int x = 1; x <= size(); x++) {
      for (int y = 1; y <= size(); y++) 
      {
        Point p = new Point (x, y);
        GoPoint gp = get_point (p);

        if (! gp.color_is(GoPoint.EMPTY))
        {
          if (fill_from_point (gp, gp.color(), base_gid))
            base_gid++;
        }
      }
    }
  }

  /**  Returns true if the board changed. */
  public boolean move(Point p)
  {
    GoPoint pt = get_point(p);
    
    // If we can't even move here, just return false.
    if (! pt.color_is(GoPoint.EMPTY))
      return false;

    // Else...

    /* The elegant algorithm is: after placing a stone, remove all
     * groups of the other color that have no liberties.  If none of
     * those got removed, then see if any groups of this color have
     * no liberties.  If any are in this circumstance, then the last
     * move was illegal.  This handles situations where the stone has
     * no liberties at the moment of placement, but creates its own
     * liberties by capturing the opponent's stones.
     *
     * Note also that some rulesets allow suicide moves, since they
     * can actually change the course of the game by opening up a Ko
     * threat.  I (kff) happen to like this way of playing, but most
     * people seem to prefer otherwise, so we'll make the default be
     * no-suicide, and someday code in the other option.  It should be
     * easy.
     */

    pt.place_stone(to_move);  // fill in the color of this GoPoint
    establish_groups();

    // Was it legal?
    if ((! remove_dead_groups(GoPoint.other_color(pt.color())))
        && (dead_groups_p(pt.color())))
    {
      pt.remove_stone(); // revert GoPoint to previous state
      return false; // because the board has NOT changed
    }

    // Else this move was legal, whether it captured anything or not:

    // Make it happen on the board.
    last_point = pt;
    to_move = GoPoint.other_color(to_move);

    // Make it "have happened" in the GameTree history:
    // ADD an SGFnode to GameTree!!
    SGFnode new_node = new SGFnode(lastNode);
    new_node.set_prop_literal(sgf_color(pt), to_sgf_coordinate(pt));
    new_node.set_prop_literal("MN", sgf_nextMoveNumber(lastNode));
    lastNode = new_node;

    // Let caller know the board has changed.
    return true;
  }

  public int liberties(Point p)
  {
    GoPoint pt;
    int count = 0;

    if ((pt = get_point(p, GoBoard.UP)) != null)
      if (pt.color_is(GoPoint.EMPTY))
        count++;
             
    if ((pt = get_point(p, GoBoard.DOWN)) != null)
      if (pt.color_is(GoPoint.EMPTY))
        count++;
             
    if ((pt = get_point(p, GoBoard.LEFT)) != null)
      if (pt.color_is(GoPoint.EMPTY))
        count++;
             
    if ((pt = get_point(p, GoBoard.RIGHT)) != null)
      if (pt.color_is(GoPoint.EMPTY))
        count++;
    
    return count;
  }
  
  // Turning GoBoard into a string is nothing more than returning the
  // string-form of our root GameTree node!

  public String toString()
  {
    return GameTree.toString();
  }

  
  public void pass() 
  {
    // Record it it the GameTree
    SGFnode new_node = new SGFnode(lastNode);
    new_node.set_prop_literal(sgf_color(to_move), "");
    new_node.set_prop_literal("MN", sgf_nextMoveNumber(lastNode));
    lastNode = new_node;
    
    // then switch the turn
    to_move = GoPoint.other_color(to_move);
  }

  public void add_comment_node(String comment)
  {
    SGFnode new_node = new SGFnode(lastNode);
    new_node.set_prop_safely("C", comment);
    lastNode = new_node;
  }

  public void score() {
    // does nothing yet
  }
  
  // return strings suitable for use in SGF properties
  
  public String sgf_color(GoPoint pt)
  {
    if (pt.color_is(GoPoint.BLACK)) {
      return "B";
    }
    if (pt.color_is(GoPoint.WHITE)) {
      return "W";
    }
    return "FOO";
  }

  public String sgf_color(int color)
  {
    if (color == GoPoint.BLACK) {
      return "B";
    }
    if (color == GoPoint.WHITE) {
      return "W";
    }
    return "FOO";
  }
  
  public String to_sgf_coordinate(GoPoint pt)
  {
    // convert GoPoint's (x,y) into an SGF letter pair.  ugh.
    return (letters[pt.x - 1] + letters[pt.y - 1]);
  }

  public int from_sgf_coordinate(char ch)
  {
    // convert an SGF alphabetic letter into an integer in GoBoard's
    // coordinate system.
    for (int i = 0; i < 52; i++) {
      if (ch == letters[i].charAt(0))
	return i+1;
    }
    return 1000; // this would be a big fat error
  }

  
  public String sgf_nextMoveNumber(SGFnode previousNode)
  {
    SGFnode tempnode = previousNode;
    String oldNumberString = null;

    // hunt up the tree (through parents) looking for the last MN[]
    // property recorded:
    while (oldNumberString == null) {
      oldNumberString = (String) tempnode.get_prop("MN");
      tempnode = tempnode.get_parent();
      if (tempnode == null) { // if we get to the root and *still* no MN
	oldNumberString = "0";
      }
    }

    // and now create a string which is one more than that
    int oldNumber = Integer.parseInt(oldNumberString);
    int nextNumber = oldNumber + 1;
    return Integer.toString(nextNumber);
  }


  // This is where we decide exactly what SGF codes we will do anything about.
  // (Ignore the ones that either don't do anything or we don't understand)

  public void execute_property(String key, String val)
  {
    if (key.equals("B")) {
      to_move = GoPoint.BLACK;
      if (val.length() == 0) {
	to_move = GoPoint.other_color(to_move); // just pass!
	return;
      }
      // else convert val to coordinates, make a black move
      int x = from_sgf_coordinate(val.charAt(0));
      int y = from_sgf_coordinate(val.charAt(1));
      Point movepoint = new Point(x, y);
      move(movepoint); // make the move
    }
    else if (key.equals("W")) {
      to_move = GoPoint.WHITE;
      if (val.length() == 0) {
	to_move = GoPoint.other_color(to_move); // just pass!
	return;
      }
      // else convert val to coordinates, make a white move
      int x = from_sgf_coordinate(val.charAt(0));
      int y = from_sgf_coordinate(val.charAt(1));
      Point movepoint = new Point(x, y);
      move(movepoint);
    }
    else if (key.equals("C")) {
      // append to our chat window (iff we belong to a GoGame)
      if (game != null) {
	// Construct a StringBuffer that has *no* escape sequences
	StringBuffer sbuf = new StringBuffer();
	for (int i = 0; i < val.length() ; i++) {
	  if (val.charAt(i) == '\\')
	    i++;
	  sbuf.append(val.charAt(i));
	}
	game.controls.chatpanel.opponent_append(sbuf.toString());
      }
    }
    else if (key.equals("PW")) {
      whitePlayerName = val;
    }
    else if (key.equals("PB")) {
      blackPlayerName = val;
    }
    else if (key.equals("SZ")) {
      // yowza!  if we get this code, it better be in the root node
      // before any moves are placed.  otherwise we're looking at some
      // big crap-ola!
      //  .....so, let's reinit our board as a new size:
      this.size(Integer.parseInt(val));
    }
    else {
      // just ignore all other property names!
    }
  }

  
}


//////////////////////////////////////////////////////////
// Don't think, just hit `y':
//
// local variables:
// eval: (load-file "bkgomagic.el")
// end:
//////////////////////////////////////////////////////////
