/** 
   The GoGame object provides the main game-window and represents a
   particular player; it collects the following classes together:

    GoPlayer -- information about the player
    GoBoard  -- the abstract model of the game in progress
    GoCanvas -- a canvas which draws the model
    BoardControls -- widgets the window's bottom (including a ChatPanel)
    MessagePasser -- talks to the bkgo server

 */

import java.io.*;
import java.net.*;
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.util.*;


public class GoGame extends Frame
{
  // Group these fields *logically*, darn it!
  boolean applet_p;          // is this class running within an applet?
  bkgo_applet parent_applet; // if running within applet, this is it!
  GoVersion version;
  
  // The component pieces which make up a "whole" bkgo client:
  GoBoard board;
  BoardControls controls;
  GoCanvas canvas;
  GoMenu menu;
  ScorePanel scorepanel;
  GoPlayer player;
  MessagePasser mp;
  String network_role;

  // Extra panels for the debug and player windows
  Panel gamepanel, userpanel;
  List userlist;
  Frame debugframe, userframe, SGFframe;
  TextArea debugtextarea, SGFtextarea;
  Container contentpane;

  // Variables & constants
  public static final int OBSERVER = 0;
  public static final int BLACK = 1;
  public static final int WHITE = 2;
  public Font bk_font, bk_font_small; // font to use everywhere we can!
  
  // For logging an error file
  boolean errorlog_open = false; 
  FileOutputStream erroroutputstream;
  PrintWriter errorlog;


  /**  Constructor creates and assembles all the parts! */

  public GoGame(boolean applet_p, bkgo_applet parent_applet)
  {
    this.applet_p = applet_p;  // should we obey applet restrictions?
    this.parent_applet = parent_applet; // our parent applet, if any
    version = new GoVersion(); 
    bk_font = new Font("SansSerif", Font.PLAIN, 12);
    bk_font_small = new Font("SansSerif", Font.PLAIN, 10);

    board = new GoBoard(this, 19);  // create a logical board, default size 19
    canvas = new GoCanvas(board, this);  // create a GoCanvas to draw model
    controls = new BoardControls(this, canvas, board); // and button-area
    scorepanel = new ScorePanel(this);
    menu = new GoMenu(this);
    
    setMenuBar(menu);

    gamepanel = new Panel();
    gamepanel.setLayout(new BorderLayout());
    gamepanel.add(scorepanel,  "North");
    gamepanel.add(canvas, "Center");
    gamepanel.add(controls, "East");
    
    add(gamepanel);

    userlist = new List(10);
    userlist.setBackground(Color.white);
    userlist.setForeground(Color.black);
    userpanel = new Panel();
    BorderLayout userlayout = new BorderLayout();
    userpanel.setLayout(userlayout);
    userpanel.add(new Label("Players Connected to GoServer:"), 
		  "North");
    userpanel.add(userlist, "Center");
    userpanel.add(new Label("BKGo Client version " + version.version_number),
                  "South");
    userframe = new Frame("BKGo: Players");
    userframe.setLayout(new BorderLayout());
    userframe.add(userpanel, "Center");
    userframe.pack();
    userframe.setSize(450, 200);


    // Also create our GoGame debug window and errorlog

    debugframe = new Frame("BKGo: Debug Messages");
    debugframe.setLayout(new BorderLayout());
    debugtextarea = new TextArea("", 15, 30); 
    debugtextarea.setEditable(false);
    debugtextarea.setForeground(Color.red);
    debugframe.add(debugtextarea, "Center");
    debugframe.pack();
    init_debug();  // start the debug file-logging

    // Also create our SGF window

    SGFframe = new Frame("BKGo: SGF Tree");
    SGFtextarea = new TextArea("", 15, 30); 
    SGFtextarea.setEditable(false);
    SGFtextarea.setForeground(Color.blue);
    SGFframe.add(SGFtextarea);
    SGFframe.pack();
    
    // The GoGame listens to a single window event -- its own close button
    this.addWindowListener( new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        quit();
      }
    });

    // But allow our debugframe and userframe to be closed as well
    debugframe.addWindowListener( new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        debugframe.setVisible(false);
	menu.debugitem.setLabel("Show debug window");
      }
    });

    SGFframe.addWindowListener( new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        SGFframe.setVisible(false);
	menu.SGFitem.setLabel("Show SGF window");
      }
    });

    userframe.addWindowListener( new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        userframe.setVisible(false);
	menu.playeritem.setLabel("Show player window");
      }
    });

    if (! applet_p) {
      setIconImage(Toolkit.getDefaultToolkit().getImage("bkgo.jpg"));
    }

    // Finally, display yourself!

    this.pack();
    this.show();
  }




  /**  Re-creates all "board" object dependencies with a GoGame;
       Called after receiving a new GoBoard over the network. */

  public void re_init(GoBoard new_board) 
  {
    board = new_board;
    board.game = this;
    canvas.board = new_board; // give canvas a new board to paint
    canvas.previous_point = null;
    controls.board = new_board;
    canvas.repaint();
    display_turn();
    display_tree();
  }
    
  /** Show who's turn it is by updating the ScorePanel */

  public void display_turn()
  {
    scorepanel.repaint();
  }


  /** Update our SGFtree display in the SGF window */

  public void display_tree()
  {
    if (SGFframe.isVisible())
      SGFtextarea.setText(board.toString());
  }


  /** Called by MessagePasser when it receives a move from the
      network; this is where GoGame actually parses the move-string
      and does the move! */
  
  public void received_move(String new_msg) {
    if (! new_msg.equals(null)) {

      int commapos = new_msg.lastIndexOf(","); // find the comma
      int x = Integer.parseInt(new_msg.substring(0, commapos));
      int y = Integer.parseInt(new_msg.substring(commapos+1));
      Point movepoint = new Point(x, y);
            
      if (board.move(movepoint)) {    // make the move
	canvas.update_display();
	display_turn();           // update the scorepanel
	display_tree();
      }

    }
  }


  /** Is it my turn or not?  Very useful routine!! If our
      network_role is DEMO, then it's *always* our turn... and thus we
      can *always* move! */

  public boolean my_turn_p() 
  {
    if (network_role.equals("DEMO"))
      return true;

    debug("Trying to place " + player.color + " stone.");
    if (player.color == board.to_move())
      return true;
    else return false;
  }    

  /** Save the GoBoard to a file. */
  public void save_board (String filename) {
    save_board_to_file(filename);
  }

  /** Helper for save_board(). */
  private void save_board_to_file(String filename)
  {
    debug("GG: Saving game to disk as " + filename);
    try
    {
      FileOutputStream fout = new FileOutputStream(filename);
      String s = board.toString();
      byte[] bytes;
      bytes = s.getBytes(); // the new Java 1.1 way to do it!
      fout.write(bytes);
      fout.flush();
      fout.close();
    }
    catch (IOException e) {
      debug("GoGame: save_board() IOException: " + e);
    }
  }

  // Load SGF file into a string, then pass this string to our parser.
  // The parser then "fills in" the children of our root node.

  public void open_board(String filename)
  {
    debug("GG: Loading game from disk: " + filename);
    SGFnode NewTree = new SGFnode();  
    StringBuffer sbuf = new StringBuffer();
    try
    {
      FileInputStream fin = new FileInputStream(filename);
      int c;
      while ((c = fin.read()) != -1) {
        sbuf.append((char) c);
      }
      fin.close();
    }
    catch (IOException e) {
      debug("GoGame.open_board():IOException: " + e);
    }

    // Fire the raw SGF text off to the server, which will then
    // automatically broadcast it back to all clients.  The clients
    // will then run their own parsers and execute the SGF commands.

    mp.send_SGF_stream(sbuf.toString());
    
  }


  /** Set up errorlog file */

  private void init_debug() {
    if (! applet_p) {
      try {
	erroroutputstream = new FileOutputStream("./errorlog");
	errorlog = new PrintWriter(erroroutputstream);
	errorlog_open = true;
	debug("(all debug messages are logged to file 'errorlog')");
      }
      catch (IOException e) {
	System.err.println("Error: can't create './errorlog'");
	errorlog_open = false;
      }
    }
  }

  /** Public routine:  allow anyone to append to our debug frame */

  public void debug(String message) {
    if (errorlog_open) {
      errorlog.println(message);
      errorlog.flush();
    }
    debugtextarea.append(message + "\n");
  }


  /** Based on sample code in the O'Reilly Tiger book: */
    
  void print_window()
  {
    // Obtain a PrintJob object, which displays a print dialog.
    Toolkit toolkit = this.getToolkit();
    PrintJob job = toolkit.getPrintJob(this, "BKGo Board", 
                                       new Properties());

    if (job == null) return; // if user cancelled
	
    Graphics page = job.getGraphics();

    // Center output on page
    Dimension size = this.getSize();
    Dimension pagesize = job.getPageDimension();
    page.translate((pagesize.width - size.width)/2,
                   (pagesize.height - size.height)/2);
    page.setClip(0,0, size.width, size.height);
	
    // All AWT components have a print() routine, which just
    // calls the paint() on a graphics object (in this case, our "page")
    gamepanel.printAll(page);

    // Clean up and send job to printer
    page.dispose();
    job.end();
  }
   
 
  /** Gracefully shut down everything to quit! */

  public void quit() 
  {
    mp.send_message("Quitting: " + player.IDnum);
    mp.quit();  // close all messagepasser streams and stop thread

    if (errorlog_open)
      errorlog.close();

    debugframe.dispose();
    userframe.dispose();
    SGFframe.dispose();
    this.dispose(); // Kill myself.
  }

  /** Same as quit(), but doesn't kill MessagePasser;  this allows
      MessagePasser to call this function first, *then* kill itself. */

  public void quit_no_mp() 
  {
    if (errorlog_open)
      errorlog.close();

    debugframe.dispose();
    userframe.dispose();
    SGFframe.dispose();
    this.dispose(); // Kill myself.
  }

}




//////////////////////////////////////////////////////////
// Don't think, just hit `y':
//
// local variables:
// eval: (load-file "bkgomagic.el")
// end:
//////////////////////////////////////////////////////////
