/**  
    This class goes hand-in-hand with a GoGame object... acting as
    it's communication channel to great world of the MessageServer and
    ClientConnectors.
*/

import java.io.*;
import java.util.*;
import java.net.*;
import java.awt.*;
import java.awt.event.*;


public class MessagePasser extends Thread {

  /** Each instance has two buffered streams for communication */

  private GoGame game;
  private DataInputStream my_input;
  private PrintWriter my_output;
  private Socket my_sock;

  /** Connects to hostname:port, sets up connected socket-streams */

  public MessagePasser (GoGame superior, String hostname, int port) 
    throws IOException {
    this.super();
    this.game = superior;
    game.debug
      ("Connecting to " + hostname + " on port " + port + "...\n");
    my_sock = new Socket(hostname, port);
    game.debug("Connection established!\n");
    my_input = new DataInputStream(my_sock.getInputStream());
    my_output = new PrintWriter(my_sock.getOutputStream());

    // Do proper handshake with the server
    String reply;
    reply = readLine(my_input); // Read server's version-string
    game.debug("Server says: " + reply);
    send_message("BKGO-Client-version-" + game.version.version_number);
    // Someday, we'll do protocol negotiations here!

    reply = readLine(my_input); // wait for OK message
    if (! reply.startsWith("OK")) {
      // protocol error, kill self.
      GoDialog gd = new GoDialog(GoDialog.INFORMATIONAL, "Sorry");
      gd.showInformationalDialog("Error: server replied: " + reply);
      game.quit_no_mp();
      quit();
    }
    else
      send_message("join");
  }


  public void quit()
  {
    try {
      my_input.close();
      my_output.close();
      my_sock.close();
      this.stop();
    }
    catch (IOException e) {
      game.debug("MP: IOException on quit(): " + e);
    }
  }

  /** Writes a line to the socket-stream;  called from the outside. */

  public void send_message(String message) 
  {
    my_output.println(message);
    my_output.flush();
  }

  /** Sends out our client's GoBoard to the server & beyond. */

  public void send_board(GoBoard board)
  {
    send_message("SendingBoard"); // tell 'em it's coming
    game.debug("MP: About to send GoBoard...");
    String SGFmessage = board.toString();
    send_message(SGFmessage);
    send_message("EOF");
    game.debug("MP: Sent GoBoard.");
  }

  public void send_SGF_stream(String SGFmessage)
  {
    send_message("SendingBoard"); // tell 'em it's coming
    game.debug("MP: About to send GoBoard...");
    send_message(SGFmessage);
    send_message("EOF");
    game.debug("MP: Sent GoBoard.");
  }

  /** readLine():  read a text string until end-of-line

      Thanks to Kevin Kelley <kelley@ruralnet.net> for this function.
      The only non-deprecated readLine() seems to be in BufferedReader,
      which royally messes up our stream when doing serialization.
      Hence, a home-made readLine():

      Three modifications from Ben:  
      1.  Deal with \r\n sent from Win32 machines (instead of just \n)
      2.  Exception handling:  kill self if network error
      3.  Set an upper limit on the size of a string received!
  */

  public String readLine(InputStream in)
  {
    StringBuffer buffer = new StringBuffer(1000); // upper limit!
    char ch = 0;
    try {
      while ((ch = (char) in.read()) != -1) { // break on EOF
        if ((ch == '\n') || (ch == '\r')) {
          if (ch == '\r') // if line ends with carriage return,
            in.read();  // read trailing newline and discard it
          break;
        }
        if (buffer.length() > 995) {
          game.debug
            ("MP: readLine(): received > 995 chars. Killing self!");
          game.quit();
        }
        else 
          buffer.append(ch);
      }
      // return null to indicate EOF; otherwise return the line.
      if (ch == -1 && buffer.length() == 0) 
        return null;
      else
        return buffer.toString();
    }
    catch (IOException e) {
      game.debug("MP: readLine() IOException,  Killing self.");
      game.quit();
    }
    catch (OutOfMemoryError e) {
      game.debug("MP: readLine OutOfMemory.  Killing self.");
      game.quit();
    }
    return null;
  }



  /**  Overriding Thread's run() method;  called when object is start()ed.*/

  public void run() {

    while (true) {  // do this forever
      
      // wait for a message indefinitely
      String message = readLine(my_input);
      
      if (message == null) {
          game.debug("Error:  received garbage/unknown message.");	  
      }

      if (message.startsWith("Message: ")) { // a chat message
	String real_message = message.substring(9); 
	game.board.add_comment_node(real_message);
	game.controls.chatpanel.opponent_append(real_message);
	game.display_tree();
      }
      
      else if (message.startsWith("Place: ")) {  // a move in the form "x,y"
	String real_message = message.substring(7);
	game.debug("Somebody placed a stone at " + real_message);
	// Send the move to our GoGame for processing:
	game.received_move(real_message);
      }
      
      else if (message.startsWith("Pass")) { // player passes turn
	game.board.pass();
	game.display_tree();
	game.controls.DisplayMessage("Player passed.");
      }
      
      else if (message.startsWith("Board: ")) { // receive a new GoBoard
	game.debug("MP: About to receive board...");
	game.controls.DisplayMessage("PLEASE WAIT: Receiving new board!");
	
	GoBoard new_gb = new GoBoard(game, 19);
	game.re_init(new_gb); // update everything!
	String message_part = null;
	StringBuffer SGFbuf = new StringBuffer();
	SGFnode NewTree = new SGFnode();
	
	message_part = readLine(my_input);
	// Keep reading lines from the SGF stream
	while (! message_part.equals("EOF")) {
	  SGFbuf.append(message_part);
	  message_part = readLine(my_input);
	}
	
	// Call the parser to build our tree
	NewTree.fromString(SGFbuf.toString());
	// Install our new blank board
	game.board.GameTree = NewTree;
	// Execute the new tree
	NewTree.execute(game.board);
	// Display it in our SGF window
	game.display_tree();
	
	game.debug("MP: Received board of size: " + 
		   new_gb.size());
	game.controls.DisplayMessage("New board received.");
      }
      
      else if (message.startsWith("GimmeGoPlayer")) { // send your GoPlayer
	game.debug("MP: About to send GoPlayer...");
	send_message("Player: " + game.player.toString());
	game.debug("MP: Sent GoPlayer.");
      }
      
      else if (message.startsWith("GimmeBoard")) { // send your GoBoard
	send_board(game.board);
      }
      
      else if (message.startsWith("GimmePassword")) { // return a password
	String password;
	if (game.network_role.equals("DEMO")) {
	  password = "demo";
	}
	else {
	  String colstring = "nobody";
	  if (game.player.color == GoPlayer.BLACK) colstring = "black's";
	  if (game.player.color == GoPlayer.WHITE) colstring = "white's";
	  GoDialog gd = new GoDialog(GoDialog.PASSWORD, "Enter Password");
	  password = gd.showPasswordDialog
	    ("The server requests " + colstring + " password:");
	}
	send_message("Password: " + password);
      }
      
      else if (message.startsWith("GimmeNewPassword")) { // set a password
	String password;
	if (game.network_role.equals("DEMO")) {
	  password = "demo";
	}
	else {
	  String colstring = "nobody";
	  if (game.player.color == GoPlayer.BLACK) colstring = "black";
	  if (game.player.color == GoPlayer.WHITE) colstring = "white";
	  GoDialog gd = new GoDialog(GoDialog.PASSWORD, "Enter Password");
	  password = gd.showPasswordDialog
	    ("The server asks you to create a password for the " 
	     + colstring + " server-owner:");
	}
	send_message("Password: " + password);
      }
      
      else if (message.startsWith("ServerRejectedFatal")) { 
	GoDialog gd = new GoDialog(GoDialog.INFORMATIONAL, "Sorry");
	gd.showInformationalDialog("The server has rejected you.");
	game.quit_no_mp();
	quit();
      }
      
      else if (message.startsWith("ServerRejectedNonFatal")) { 
	GoDialog gd = new GoDialog(GoDialog.INFORMATIONAL, "Sorry");
	gd.showInformationalDialog("The server has denied the password.");
      }
      
      else if (message.startsWith("ClearPlayers")) { // clear userlist
	// Empty out our vector of playerstrings.
	game.userlist.removeAll();
      }
      
      else if (message.startsWith("AddPlayer: ")) { // append to userlist
	String new_name = message.substring(11);
	game.userlist.add(new_name);
      }
      
      else if (message.startsWith("BlackPlayer: ")) { // set name in GoBoard
	String new_name = message.substring(13);
	game.board.set_blackPlayerName(new_name);
      }
      
      else if (message.startsWith("WhitePlayer: ")) { // set name in GoBoard
	String new_name = message.substring(13);
	game.board.set_whitePlayerName(new_name);
      }
      
      else if (message.startsWith("AnnouncePlayer: ")) { // new arrival
	String new_name = message.substring(16);
	game.controls.chatpanel.opponent_append
	  ("\n***> New Player Has Arrived: " + new_name + " <***\n");
      }
      
      else if (message.startsWith("GoodbyePlayer: ")) { // new departure
          String new_name = message.substring(15);
          game.controls.chatpanel.opponent_append
	    ("\n***> Player Has Departed: " + new_name + " <***\n");
      }
      
      else if (message.startsWith("Dialog: ")) { 
	String announcement = message.substring(8);
	GoDialog gd = new GoDialog
	  (GoDialog.INFORMATIONAL, "Message from Server");
	gd.showInformationalDialog(announcement);
      }
      
      else if (message.startsWith("ID: ")) { // receive your unique ID num
	String identity = message.substring(4);
	int ID = Integer.parseInt(identity);
	game.player.IDnum = ID;
	game.debug("MP: Server says my unique ID is " + ID);
      }
      
      else {
	game.debug("Error:  received garbage/unknown message.");
      }

    }
  }

}


//////////////////////////////////////////////////////////
// Don't think, just hit `y':
//
// local variables:
// eval: (load-file "bkgomagic.el")
// end:
//////////////////////////////////////////////////////////
