/**  
    MessageServer class:  the basic bkgo server.

*/

import java.io.*;
import java.util.*;
import java.net.*;
import java.awt.*;
import java.awt.event.*;


public class MessageServer extends Thread {

  GoVersion version;            // what version is this server?
  ServerSocket my_ss;           // serversocket to listen on
  private int port;             // the portnumber I'm listening on
  Vector players;               // list of currently connected GoPlayers
  GoBoard board;                // my global model of the game in progress
  String blackpassword;         // password of the black server-owner
  String whitepassword;         // password of the white server-owner
  private int ID_count = 0;     // assign unique IDs to clients
  String myname;                // "name" given to this server by players
  String filename;              // reflect connected players in this file
  PrintWriter statuswriter;     // used for writing to the file


  public MessageServer(int port) throws IOException {
    this.super();
    this.version = new GoVersion();
    this.port = port;
    this.players = new Vector();
    this.my_ss = new ServerSocket(port);
    this.board = new GoBoard(19);         // make a blank board
    this.blackpassword = "none";          // this server has no owner at start
    this.whitepassword = "none";
    this.myname = "none";

    filename = "./goserver_status." + port;
  }

  /** Graceful termination */

  public void quit()
  {
    try {
      my_ss.close();
      File killme = new File(filename);
      killme.delete();
      System.exit(0);
    }
    catch (IOException e) {
      System.out.println("MS: Can't close my serversocket: " + e);
      this.stop();
    }
  }


  /**  Broadcast a general message to all our ClientConnectors */
    
  public void broadcast(String message) 
  {
    for (int i = 0; i < players.size(); i++) {
      GoPlayer gp = (GoPlayer) players.elementAt(i);
      gp.CC.send_message(message);
    }
  }

  /**  Broadcast the list of players to all our ClientConnectors.

       If (join == true), then we announce the arrival of playername.
       Otherwise, we announce this player's departure!
   */

  public void broadcast_players(String playername, int ID, boolean join) 
  {
    GoPlayer gp, gp2;
    String info = null;
    Date current_date = Calendar.getInstance().getTime();
    String build_list = current_date.toString() + ":\n";

    System.out.println("MS:  Sending new playerlist out to all...");
    for (int i = 0; i < players.size(); i++)  // who we're sending to
    {
      gp = (GoPlayer) players.elementAt(i);
      gp.CC.send_message("ClearPlayers"); // clear their lists
      info = gp.toPrettyString();
      build_list = build_list + info + "\n";

      for (int j = 0; j < players.size(); j++) // list all players
      { 
        gp2 = (GoPlayer) players.elementAt(j);
        gp.CC.send_message("AddPlayer: " + gp2.toPrettyString());
	if (gp2.color == GoPlayer.BLACK) {
	  gp.CC.send_message("BlackPlayer: " + gp2.name);
	}
	if (gp2.color == GoPlayer.WHITE) {
	  gp.CC.send_message("WhitePlayer: " + gp2.name);
	}
      }
      if (join && (ID != (gp.IDnum))) {
	gp.CC.send_message("AnnouncePlayer: " + playername);
      }
      else if (!join) {
	gp.CC.send_message("GoodbyePlayer: " + playername);
      }
    } 
    try {
      statuswriter = new PrintWriter(new FileWriter(filename, false), true);
      statuswriter.println(build_list);
    }
    catch (IOException e) {
      System.out.println("MS: Can't output to " + filename + ":" + e);
    }
  }

  /** Broadcast a new board to all our players */

  public void broadcast_board(GoBoard new_board)
  {
    System.out.println("MS:  Sending new GoBoard out to all...");
    for (int i = 0; i < players.size(); i++) {
      GoPlayer gp = (GoPlayer) players.elementAt(i);
      gp.CC.send_board(new_board);
    }
  }


  /** Either set the server-owner passwords or check against them! 
      Returns boolean success.
  */

  public boolean authenticate_player(GoPlayer player)
  {
    if ( (player.color == GoPlayer.BLACK) && 
	 (blackpassword.equals("none")) )
    {
      // This is the first time anyone claims to be "black", so 
      // congrats:  you become a co-owner of this server!
      
      // query user for blackpassword here and set it.
      blackpassword = player.CC.fetch_password(true);
      return true;
    }

    if ( (player.color == GoPlayer.WHITE) && 
	 (whitepassword.equals("none")) )
    {
      // This is the first time anyone claims to be "white", so 
      // congrats:  you become a co-owner of this server!
      
      // query user for whitepassword here and set it.
      whitepassword = player.CC.fetch_password(true);
      return true;
    }

    // If we get here, then this server already has owner(s).

    if (player.color == GoPlayer.BLACK)
    {
      String testpassword = player.CC.fetch_password(false);
      if (! testpassword.startsWith(blackpassword))
      {
	System.out.println("MS: Received incorrect password!");
	return false;
      }
      else return true;
    }

    if (player.color == GoPlayer.WHITE)
    {
      String testpassword = player.CC.fetch_password(false);
      if (! testpassword.startsWith(whitepassword))
      {
	System.out.println("MS: Received incorrect password!");
	return false;
      }
      else return true;
    }


    return false;
  }



  /**  Add a new GoPlayer to my Vector, then re-broadcast GoBoard to all  */

  public void add_player(ClientConnector cc)
  {
    // Fill in this missing field in GoPlayer object
    cc.my_player.CC = cc;

    // Authenticate them!
    if (cc.my_player.color != GoPlayer.OBSERVER) // observers are always OK
    {
      if (! authenticate_player(cc.my_player)) // but b&w aren't...
      {
	cc.send_message("ServerRejectedFatal");
	return;
      }
    }

    // Assuming the player is authenticated, 

    // Send him/her a unique ID number
    ID_count++;
    cc.my_player.IDnum = ID_count;
    cc.send_message("ID: " + ID_count);

    // add the GoPlayer to our list of currently connected players
    players.addElement(cc.my_player);

    // set the name of the player in our own GoBoard
    if (cc.my_player.color == GoPlayer.BLACK) {
      board.set_blackPlayerName(cc.my_player.name);
    }
    if (cc.my_player.color == GoPlayer.WHITE) {
      board.set_whitePlayerName(cc.my_player.name);
    }

    // send this new player the latest GoBoard
    cc.send_board(board);

    // and broadcast the new list of players to everybody.
    System.out.println
      ("MS: New player connected; my playerlist is now size "
       + players.size() + ":");
    broadcast_players(cc.my_player.toPrettyString(), cc.my_player.IDnum, true);
  }



  /** Remove a player from the game */

  public void remove_player(int ID)
  {
    GoPlayer dead_player = null;
    boolean do_remove = false;

    System.out.println("MS:  I am attempting to remove a player!");
    for (int i = 0; i < players.size(); i++) {
      GoPlayer gp = (GoPlayer) players.elementAt(i);
      System.out.println("MS: Searching for ID " + ID + 
                 ", found: ID " + gp.IDnum);	   
      if (ID == gp.IDnum) {
        System.out.println("MS: Found player to remove!");
        dead_player = gp;
        do_remove = true;
      }
    }	
    if (do_remove) {
      String dead_description = dead_player.toPrettyString();
      int dead_ID = dead_player.IDnum;
      players.removeElement(dead_player);
      System.out.println
        ("MS: My playerlist is now size "
         + players.size() + ":");
      broadcast_players(dead_description, dead_ID, false);
    }
  }
  


  public void received_move(String coordinates) {
    if (! coordinates.equals(null)) 
    {
      int commapos = coordinates.lastIndexOf(","); // find the comma
      int x = Integer.parseInt(coordinates.substring(0, commapos));
      int y = Integer.parseInt(coordinates.substring(commapos+1));
      Point movepoint = new Point(x, y);
      
      if( board.move(movepoint))  // place the stone
      {   
	// if successful in placing the stone,
	// save board to disk?
	save_board_to_file("./game_in_progress." + port);
      }
    }
  }


  private void save_board_to_file(String filename)
  {
    try
    {
      FileOutputStream fout = new FileOutputStream(filename);
      String s = board.toString();
      byte[] bytes;
      bytes = s.getBytes(); // the new Java 1.1 way to do it!
      fout.write(bytes);
      fout.flush();
      fout.close();
    }
    catch (IOException e) {
      System.out.println("MS: save_board_to_file() IOException: " + e);
    }
  }

  /** Overriding Thread's run() method;  called when object is start()ed. 

      MessageServer's main thread forever blocks on the ServerSocket,
      waiting for new clients to connect.
  */
    
  public void run() {
    while (true) { 
      try {
	System.out.println("MS: Waiting for new connections...");
        // block indefinitely, waiting for new socket connections
        Socket new_socket = my_ss.accept();
        // when you get one, create a ClientConnector around it
        ClientConnector new_client = 
          new ClientConnector(new_socket, this, ID_count);
	System.out.println("MS: Got new connection, doing handshake...");
	// try to handshake with the client
	if (! new_client.do_handshake()) 
	{
	  // handshake failed, so...
	  new_client.die(false);
	}
	else { // handshake ok, so start listening...
	  new_client.start();
	}
      }
      catch (IOException e) {
        System.err.println("IOException on MessageServer run(): " + e);
      }
    }
  }


}


//////////////////////////////////////////////////////////
// Don't think, just hit `y':
//
// local variables:
// eval: (load-file "bkgomagic.el")
// end:
//////////////////////////////////////////////////////////
